/*** machine.h  -*- C -*- */
/** Atmel AVR backend for small machine abstraction library. */

/*** Ivan Shmakov, 2019, 2020, 2022 */

/** To the extent possible under law, the author(s) have dedicated
 ** all copyright and related and neighboring rights to this software
 ** to the public domain worldwide.  This software is distributed
 ** without any warranty.
 **
 ** You should have received a copy of the CC0 Public Domain Dedication
 ** along with this software.  If not, see
 ** <http://creativecommons.org/publicdomain/zero/1.0/>.
 */

/*** History: */

/** 0.? 2022-10-02 20:37Z
 **     Minor wording and coding style tweaks all across the code.
 **     New FIXME notes.
 **     (MACHINE_TIMER_DIVISOR): No longer use, or define, this macro.
 **     (MACHINE_TIMER_US): Require a single divisor argument.
 **     (machine_timer_setup): Likewise.
 **     (MACHINE_UARTS): New macro, depending on F_CPU and BAUD
 **     and UDR0 (or UDR) being defined.  Use it across the code
 **     in place of the former F_CPU and BAUD being defined test.
 **     (MACHINE_MEGA8_LIKE): New macro.
 **     (MACHINE_MEGA64_LIKE): New macro.
 **     (MACHINE_TCCR0B_COMPUTE): New macro, split off
 **     (USART_RX_vect): Only define ISR if available and MACHINE_UARTS
 **     is greater than zero.
 **     (USART_TX_vect): Likewise.
 **     (USART0_RX_vect): Likewise.
 **     (USART0_TX_vect): Likewise.
 **     (machine_timer_setup): this function.  Explicitly support
 **     the likes of ATmega8, as well the likes of ATmega64.  The macro
 **     is not defined otherwise.
 **     (machine_spi_master_init): Only configure UART 0 as SPI
 **     master when the argument is 1 (was: non-zero) and only support
 **     it if UARTs are available and UMSEL01 is defined.  Explicitly
 **     support the likes of ATmega8, with /SS on PB2, and the likes
 **     of ATmega64, with /SS on PB0 (was: assuming /SS always on PB2.)
 **     (machine_spi_master_init): Wrap spi_ready_p manipulation
 **     in an ATOMIC_BLOCK, so the function can also be used when
 **     interrupts are enabled.
 **     (machine_spi_slave_init): Likewise.
 **     (machine_spi_ready_p): New name for machine_spi_ready (old one
 **     retained as a macro.)
 **     (machine_port_map_t): New type, defined as either uint_least64_t
 **     or uint_least32_t depending on the number of GPIO ports
 **     on the particular MCU.
 **     (machine_port_setup): Add support for PORTE through PORTH.
 **     (machine_port_read): Likewise; shortened through the use of
 **     a temporarily-defined MACHINE_PORT_READ macro.
 **     (machine_port_write): Likewise; mask the data value as well.
 **
 ** 0.? 2020-03-14 11:25:18Z
 **     (sfn.ogZ3YFYlfkTxSQCxorAGG_YVRXOR-nkmisFBeT7jo6k.h)
 **     Tentatively prepended channel argument to SPI (non-zero
 **     mapped to USART in SPI master mode) and UART (ignored)
 **     functions declarations.
 **     (machine_eeprom_read): New function.
 **     (machine_eeprom_write): Likewise.
 **/

/*** Code: */
#ifndef MACHINE_H
#define MACHINE_H

#include <assert.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <stddef.h>             /* for ptrdiff_t */
#include <stdint.h>
#include <util/atomic.h>


/** FIXME:  static_assert does not quite work with inline functions?
 **
 **         The idea here is that when the compiler is unable to prove
 **         that the assertion is never violated, the resulting binary
 **         will contain the respective constructed string.
 */
#define x_assert(x, y) assert ((y) && (x))

#if ! defined (F_CPU)
#warning "F_CPU is not defined; UART et all will not be available"
#else
#define MACHINE_TIMER_LIMIT     (256L)
#define MACHINE_TIMER_US(divisor) \
  (1 + (-1LL + 1000000LL * (MACHINE_TIMER_LIMIT) * (divisor)) \
   / (F_CPU))
#if defined (BAUD)
#include <util/setbaud.h>
#if defined (UDR1)
#define MACHINE_UARTS (2)
#elif defined (UDR0) || defined (UDR)
#define MACHINE_UARTS (1)
#else
#define MACHINE_UARTS (0)
#endif
#else
#define MACHINE_UARTS (0)
#warning "BAUD is not defined; UART will not be available"
#endif
#endif

#define MACHINE_MEGA8_LIKE \
        defined (__AVR_ATmega8__) || defined (__AVR_ATmega8A__) \
     || defined (__AVR_ATmega48__) || defined (__AVR_ATmega48A__) \
     || defined (__AVR_ATmega48PA__) || defined (__AVR_ATmega48PB__) \
     || defined (__AVR_ATmega88__) || defined (__AVR_ATmega88A__) \
     || defined (__AVR_ATmega88PA__) || defined (__AVR_ATmega88PB__) \
     || defined (__AVR_ATmega168__) || defined (__AVR_ATmega168A__) \
     || defined (__AVR_ATmega168PA__) || defined (__AVR_ATmega168PB__) \
     || defined (__AVR_ATmega328__) || defined (__AVR_ATmega328A__) \
     || defined (__AVR_ATmega328PA__) || defined (__AVR_ATmega328PB__)
#define MACHINE_MEGA64_LIKE \
        defined (__AVR_ATmega64__) || defined (__AVR_ATmega64A__) \
     || defined (__AVR_ATmega128__) || defined (__AVR_ATmega128A__)

/** NB: we assume that CS00.2 are 0 through 2, so N << CS0 is N */

/** Once again, should the compiler be unable to prove
 ** that only supported divisors are ever used, the resulting
 ** binary will end up containing the string below
 */
#if MACHINE_MEGA8_LIKE
#define MACHINE_TCCR0B_COMPUTE(divisor) \
    ((divisor) == 1 ? 1 : (divisor) == 8 ? 2 : (divisor) == 64 ? 3 \
     : (divisor) == 256L ? (4) : (divisor) == 1024L ? (5) \
     : (int_least8_t)"timer divisor is not one of the supported values")
#elif MACHINE_MEGA64_LIKE
#define MACHINE_TCCR0B_COMPUTE(divisor) \
    ((divisor) == 1 ? 1 : (divisor) == 8 ? 2 : (divisor) == 32 ? 3 \
     : (divisor) == 64 ? 4 : (divisor) == 128L ? 5 \
     : (divisor) == 256L ? 6 : (divisor) == 1024L ? 7 \
     : (int_least8_t)"timer divisor is not one of the supported values")
#else
#warning "Unknown MCU; machine_timer_setup will not be available"
#endif


/** Compatibility with certain older MCUs (ATmega8, ATmega64A, etc.) */
/** FIXME: may interfere with user code? */
#if (defined (TIMSK) && ! defined (TIMSK0))
#define TIMSK0  TIMSK
#endif
#if (defined (TCCR0) && ! defined (TCCR0B))
#define TCCR0B  TCCR0
#endif
#if ! defined (UBRR0H) && defined (UBRRH)
#define UDR0    UDR
#define UBRR0H  UBRRH
#define UBRR0L  UBRRL
#define UCSR0A  UCSRA
#define UCSR0B  UCSRB
#define UCSR0C  UCSRC
#define U2X0    U2X
#define UDRE0   UDRE
#define RXC0    RXC
#define TXEN0   TXEN
#define RXEN0   RXEN
#define TXCIE0  TXCIE
#define RXCIE0  RXCIE
#define UCSZ00  UCSZ0
#endif
#ifdef  URSEL
#define URSELBIT  (1 << (URSEL))
#else
#define URSELBIT  0
#endif


/** NB: generally, writes to machine_flags should be inside
 **     ATOMIC_BLOCK, as those get translated into read-modify-write
 */
volatile struct {
  unsigned  timer_p     : 1;
  unsigned  spi_ready_p : 1;
} machine_flags = { 0 };

volatile uint_least8_t machine_spi_last_read, machine_spi_last_write = 0;


ISR (TIMER0_OVF_vect)
{
  machine_flags.timer_p = 1;
}

static inline int_least8_t machine_spi_master_p (uint_least8_t);

ISR (SPI_STC_vect)
{
  machine_flags.spi_ready_p = 1;

  /** NB: a kind of a buffer implemented in software */
  machine_spi_last_read = SPDR;
  if (! machine_spi_master_p (0)) {
    SPDR = machine_spi_last_write;
  }
}

#if MACHINE_UARTS > 0
#if defined (USART_RX_vect)
ISR (USART_RX_vect)
{
  /* do nothing */
}

ISR (USART_TX_vect)
{
  /* do nothing */
}
#endif

#if defined (USART0_RX_vect)
ISR (USART0_RX_vect)
{
  /* do nothing */
}

ISR (USART0_TX_vect)
{
  /* do nothing */
}
#endif
#endif


/** FIXME: separate machine_wdt_setup and machine_wait_setup functions? */
static inline void
machine_init (void)
{
  /** FIXME: hardcoded 2 s watchdog timer interval */
  wdt_enable (WDTO_2S);
  sleep_enable ();
}

static inline void
machine_wait (void)
{
  /** NB: we only ever explicitly enable interrupts here */
  sei ();
  wdt_reset ();
  sleep_cpu ();
}


#ifdef MACHINE_TCCR0B_COMPUTE
static inline void
machine_timer_setup (uint_least16_t divisor)
{
  TCCR0B  = MACHINE_TCCR0B_COMPUTE (divisor);
  TIMSK0 |= (1 << TOIE0); /* timer 0 overflow */
}
#endif

static inline int_least8_t
machine_timer_ticked_p (void)
{
  int_least8_t r;
  ATOMIC_BLOCK (ATOMIC_RESTORESTATE) {
    r = machine_flags.timer_p;
    machine_flags.timer_p
      = 0;
  }
  /* . */
  return r;
}


static inline uint_least8_t
machine_eeprom_read (ptrdiff_t address)
{
  /* . */
  return
    eeprom_read_byte ((const uint8_t *)address);
}

static inline void
machine_eeprom_write (ptrdiff_t address, uint_least8_t v)
{
  eeprom_write_byte ((uint8_t *)address, v);
  /* . */
}


#if MACHINE_UARTS > 0
static inline void machine_uart_setup (uint_least8_t);
static inline int_least8_t machine_uart_can_read_p (uint_least8_t);
static inline int_least8_t machine_uart_can_write_p (uint_least8_t);
static inline int_least16_t machine_uart_read (uint_least8_t);
static inline void machine_uart_write (uint_least8_t, uint_least8_t);
#endif


static inline int_least8_t
machine_spi_master_p (uint_least8_t cc)
{
  /* . */
  return (cc ? 1 : ((1 << MSTR) & SPCR));
}

static inline void
machine_spi_master_init (uint_least8_t cc)
{
#if defined (UMSEL01) && MACHINE_UARTS >= 1
  if (1 == cc) {
    /** FIXME: we assume that UBRR is 0 at this point, as needed */
    /** FIXME: allow enabling only Rx or only Tx */
#if 1
    UCSR0B = ((1   << TXEN0)  | (1 << RXEN0)
              | (1 << TXCIE0) | (1 << RXCIE0));
#else
    /** FIXME: transmit-only mode hardcoded for now */
    UCSR0B = ((1   << TXEN0)
              | (1 << TXCIE0));
#endif

    /** master SPI */
    /** FIXME: no way to configure UCPOL, UCPHA, UDORD */
    UCSR0C = (URSELBIT
              | (1 << UMSEL01) | (1 << UMSEL00)); /* master SPI mode */

    /** NB: UBRR is set the last */
    UBRR0H  = UBRRH_VALUE;
    UBRR0L  = UBRRL_VALUE;
#if USE_2X
    UCSR0A |=   (1 << U2X0);
#else
    UCSR0A &= ~ (1 << U2X0);
#endif

    /* . */
    return;
  }
#endif

  x_assert (0 == cc, "only SPI channel 0 is supported");

  /** NB: low /SS as input is used in multi-master SPI;
   **     set it up as output, 1 to ensure single-master
   */
#if MACHINE_MEGA8_LIKE
  DDRB  |= (1 <<   DDB2);
  PORTB |= (1 << PORTB2);
#elif MACHINE_MEGA64_LIKE
  DDRB  |= (1 <<   DDB0);
  PORTB |= (1 << PORTB0);
#else
#warning "Unknown MCU; machine_spi_master_init will not force /SS output, 1"
#endif
  /** FIXME: no way to configure DORD, CPOL, CPHA, SPI2X, SPR0.1 */
  SPCR = ((1 << SPIE) | (1 << SPE) | (1 << MSTR) | (1 << SPR1));
  ATOMIC_BLOCK (ATOMIC_RESTORESTATE) {
    machine_flags.spi_ready_p = 1;
  }
  /* . */
}

static inline void
machine_spi_slave_init (uint_least8_t cc)
{
  x_assert (0 == cc, "slave mode is only supported for SPI channel 0");

  /** FIXME: no way to configure DORD, CPOL, CPHA */
  SPCR = ((1 << SPIE) | (1 << SPE));
  ATOMIC_BLOCK (ATOMIC_RESTORESTATE) {
    machine_flags.spi_ready_p = 0;
  }
  /* . */
}

/** NB: compatibility alias */
#define machine_spi_ready machine_spi_ready_p

static inline int_least8_t
machine_spi_ready_p (uint_least8_t cc)
{
  if (cc) {
    /** FIXME: now this seems ad-hoc? */
    /* . */
    return machine_uart_can_write_p (0);
  }

  /* . */
  return machine_flags.spi_ready_p;
}

static inline uint_least8_t
machine_spi_read (uint_least8_t cc)
{
  if (cc) {
    /* . */
    return machine_uart_read (0);
  }

  ATOMIC_BLOCK (ATOMIC_RESTORESTATE) {
    if (! machine_spi_master_p (cc)) {
      machine_flags.spi_ready_p = 0;
    }
  }

  /* . */
  return machine_spi_last_read;
}

static inline void
machine_spi_write (uint_least8_t cc, uint_least8_t v)
{
  if (cc) {
    machine_uart_write (0, v);
    /* . */
    return;
  }

  ATOMIC_BLOCK (ATOMIC_RESTORESTATE) {
    if (machine_spi_master_p (cc)) {
      machine_flags.spi_ready_p = 0;
    }
    SPDR = machine_spi_last_write = v;
  }
  /* . */
}


#if MACHINE_UARTS > 0
/** FIXME: only USART0 is supported below; the cc argument is ignored */

static inline void
machine_uart_setup (uint_least8_t cc)
{
  UBRR0H  = UBRRH_VALUE;
  UBRR0L  = UBRRL_VALUE;
#if USE_2X
  UCSR0A |=   (1 << U2X0);
#else
  UCSR0A &= ~ (1 << U2X0);
#endif

  /** FIXME: allow enabling only Rx or only Tx */
  UCSR0B = ((1   << TXEN0)  | (1 << RXEN0)
            | (1 << TXCIE0) | (1 << RXCIE0));
  /** FIXME: hardcoded to asynchronous, 8N1 */
  UCSR0C  = URSELBIT | (3 << UCSZ00);
  /* . */
}

static inline int_least8_t
machine_uart_can_read_p (uint_least8_t cc)
{
  /* . */
  return (UCSR0A & (1 << RXC0));
}

static inline int_least8_t
machine_uart_can_write_p (uint_least8_t cc)
{
  /* . */
  return (UCSR0A & (1 << UDRE0));
}

static inline int_least16_t
machine_uart_read (uint_least8_t cc)
{
  /* . */
  return (machine_uart_can_read_p (cc) ? UDR0 : -1);
}

static inline void
machine_uart_write (uint_least8_t cc, uint_least8_t v)
{
  UDR0 = v;
  /* . */
}
#endif


/** NB: this should be the only place in this file
 **     where such wide integer types are used;
 **     we rely on compiler optimizing the access
 **     to the relevant registers to never actually
 **     use wide-integer operations on the MCU itself
 */
#if defined (PORTE) || defined (PORTF) || defined (PORTG) || defined (PORTH)
typedef uint_least64_t machine_port_map_t;
#else
typedef uint_least32_t machine_port_map_t;
#endif

static inline void
machine_port_setup (machine_port_map_t read_mask,
                    machine_port_map_t write_mask)
{
  x_assert ((read_mask & write_mask) == 0, "read and write masks clash");

#ifdef DDRA
  DDRA = (write_mask & 0377);
#else
  x_assert (((write_mask | read_mask) & 0377) == 0,
            "No port A on this chip");
#endif
  read_mask >>= 8;
  write_mask >>= 8;
#ifdef DDRB
  DDRB = (write_mask & 0377);
#else
  x_assert (((write_mask | read_mask) & 0377) == 0,
            "No port B on this chip");
#endif
  read_mask >>= 8;
  write_mask >>= 8;
#ifdef DDRC
  DDRC = (write_mask & 0377);
#else
  x_assert (((write_mask | read_mask) & 0377) == 0,
            "No port C on this chip");
#endif
  read_mask >>= 8;
  write_mask >>= 8;
#ifdef DDRD
  DDRD = (write_mask & 0377);
#else
  x_assert (((write_mask | read_mask) & 0377) == 0,
            "No port D on this chip");
#endif
  read_mask >>= 8;
  write_mask >>= 8;
#ifdef DDRE
  DDRE = (write_mask & 0377);
#else
  x_assert (((write_mask | read_mask) & 0377) == 0,
            "No port E on this chip");
#endif
  read_mask >>= 8;
  write_mask >>= 8;
#ifdef DDRF
  DDRF = (write_mask & 0377);
#else
  x_assert (((write_mask | read_mask) & 0377) == 0,
            "No port F on this chip");
#endif
  read_mask >>= 8;
  write_mask >>= 8;
#ifdef DDRG
  DDRG = (write_mask & 0377);
#else
  x_assert (((write_mask | read_mask) & 0377) == 0,
            "No port G on this chip");
#endif
  read_mask >>= 8;
  write_mask >>= 8;
#ifdef DDRH
  DDRH = (write_mask & 0377);
#else
  x_assert (((write_mask | read_mask) & 0377) == 0,
            "No port H on this chip");
#endif
  x_assert (write_mask == 0,
            "Non-zero write mask given for nonexistent port");
  x_assert (read_mask == 0,
            "Non-zero read mask given for nonexistent port");
  /* . */
}

static inline machine_port_map_t
machine_port_read  (uint_least8_t shift, machine_port_map_t mask)
{
  machine_port_map_t r = 0, mm = (mask << shift);
#define MACHINE_PORT_READ(psh, pin) \
  if (((0377ULL << psh) & mm) != 0) { \
    machine_port_map_t v = pin; \
    r |= (mask & ((v << psh) >> shift)); \
  }
#ifndef PINA
  x_assert (0 == ((0xffULL) & mm), "No port A on this chip");
#else
  MACHINE_PORT_READ (0, PINA);
#endif
#ifndef PINB
  x_assert (0 == ((0377ULL << 8) & mm), "No port B on this chip");
#else
  MACHINE_PORT_READ (8, PINB);
#endif
#ifndef PINC
  x_assert (0 == ((0377ULL << 16) & mm), "No port C on this chip");
#else
  MACHINE_PORT_READ (16, PINC);
#endif
#ifndef PIND
  x_assert (0 == ((0377ULL << 24) & mm), "No port D on this chip");
#else
  MACHINE_PORT_READ (24, PIND);
#endif
#ifndef PINE
  x_assert (0 == ((0377ULL << 32) & mm), "No port E on this chip");
#else
  MACHINE_PORT_READ (32, PINE);
#endif
#ifndef PINF
  x_assert (0 == ((0377ULL << 40) & mm), "No port F on this chip");
#else
  MACHINE_PORT_READ (40, PINF);
#endif
#ifndef PING
  x_assert (0 == ((0377ULL << 48) & mm), "No port G on this chip");
#else
  MACHINE_PORT_READ (48, PING);
#endif
#ifndef PINH
  x_assert (0 == ((0377ULL << 56) & mm), "No port H on this chip");
#else
  MACHINE_PORT_READ (56, PINH);
#endif
#undef MACHINE_PORT_READ

  /* . */
  return r;
}

static inline void
machine_port_write (machine_port_map_t data, machine_port_map_t mask)
{
#ifdef PORTA
  if ((0377 & mask)) {
    PORTA = ((PORTA & (0377 ^ mask)) | (0377 & mask & data));
  }
#else
  x_assert ((mask & 0377) == 0, "No port A on this chip");
#endif
  data >>= 8;
  mask >>= 8;
#ifdef PORTB
  if ((0377 & mask)) {
    PORTB = ((PORTB & (0377 ^ mask)) | (0377 & mask & data));
  }
#else
  x_assert ((mask & 0377) == 0, "No port B on this chip");
#endif
  data >>= 8;
  mask >>= 8;
#ifdef PORTC
  if ((0377 & mask)) {
    PORTC = ((PORTC & (0377 ^ mask)) | (0377 & mask & data));
  }
#else
  x_assert ((mask & 0377) == 0, "No port C on this chip");
#endif
  data >>= 8;
  mask >>= 8;
#ifdef PORTD
  if ((0377 & mask)) {
    PORTD = ((PORTD & (0377 ^ mask)) | (0377 & mask & data));
  }
#else
  x_assert ((mask & 0377) == 0, "No port D on this chip");
#endif
  data >>= 8;
  mask >>= 8;
#ifdef PORTE
  if ((0377 & mask)) {
    PORTE = ((PORTE & (0377 ^ mask)) | (0377 & mask & data));
  }
#else
  x_assert ((mask & 0377) == 0, "No port E on this chip");
#endif
  data >>= 8;
  mask >>= 8;
#ifdef PORTF
  if ((0377 & mask)) {
    PORTF = ((PORTF & (0377 ^ mask)) | (0377 & mask & data));
  }
#else
  x_assert ((mask & 0377) == 0, "No port F on this chip");
#endif
  data >>= 8;
  mask >>= 8;
#ifdef PORTG
  if ((0377 & mask)) {
    PORTG = ((PORTG & (0377 ^ mask)) | (0377 & mask & data));
  }
#else
  x_assert ((mask & 0377) == 0, "No port G on this chip");
#endif
  data >>= 8;
  mask >>= 8;
#ifdef PORTH
  if ((0377 & mask)) {
    PORTH = ((PORTH & (0377 ^ mask)) | (0377 & mask & data));
  }
#else
  x_assert ((mask & 0377) == 0, "No port H on this chip");
#endif

  /* . */
}

#endif
/*** machine.h ends here */
