/*** machine.c  -*- C -*- */
/** GNU backend for small machine abstraction library. */

/*** Copyright (C) 2019, 2020, 2022 Ivan Shmakov */

/** Permission to copy this software, to modify it, to redistribute it,
 ** to distribute modified versions, and to use it for any purpose is
 ** granted, subject to the following restrictions and understandings.
 **
 ** 1.  Any copy made of this software must include this copyright notice
 **     in full.
 **
 ** 2.  I have made no warranty or representation that the operation of
 **     this software will be error-free, and I am under no obligation to
 **     provide any services, by way of maintenance, update, or otherwise.
 **
 ** 3.  In conjunction with products arising from the use of this material,
 **     there shall be no use of my name in any advertising, promotional,
 **     or sales literature without prior written consent in each case.
 */

/*** History: */

/** 0.? 2022-10-04
 **     Use a BSD-like license (was: CC0.)  Minor coding style tweaks
 **     all across the code.
 **     (SIZE): Fixed: wrap the argument into parentheses under
 **     dereferencing.
 **     (MACHINE_TIMER_BASE_US_DFL): New macro, replacing
 **     (MACHINE_TIMER_US): this one.
 **     (endpoint_prefix): New unit-local variable.
 **     (tick): Likewise.
 **     (spi_master_p): Likewise.
 **     (spi_ready_p): Likewise.
 **     (endpoint_filename): Use endpoint_prefix (was: /tmp and getpid.)
 **     (endpoint_filename_alloc): New unit-local function.
 **     (spi_init): Likewise.
 **     (ensure_size): Use ftruncate (was: writing a buffer of zeros.)
 **     (new_socket): Fixed: use sigemptyset to properly initialize
 **     .sa_mask (was: not initialized); do not try to set (internal)
 **     .sa_restorer.
 **     (machine_timer_setup): Likewise.
 **     (fds_ready): Removed unused unit-local function.
 **     (fds_set): Do not accept a pointer argument and just return
 **     the highest-numbered file descriptor instead.
 **     (fds_accept): Issue a warning and close the connection just
 **     accepted if there is no room to store the file descriptor
 **     in the array provided (was: assertion violation.)  Update
 **     the calls from across the code to omit the listening socket
 **     from the vector passed to this function.
 **     (fds_setup_x): New unit-local function split off
 **     (machine_spi_slave_init): this
 **     (machine_uart_setup): and this functions.  Return early
 **     if the listening socket descriptor is already set.
 **     (fds_write): New unit-local function split off
 **     (machine_uart_write): this function.
 **     (machine_init): Set endpoint_prefix to either the copy
 **     of the value of the MACHINE_E_PREFIX environment variable,
 **     or a string constructed from TMPDIR (or /tmp) and
 **     the value returned by getpid.
 **     (machine_timer_handler): Now a unit-local function.
 **     Increment tick as well.
 **     (machine_timer_setup): Require a single divisor argument.
 **     Use the value of the MACHINE_TIMER_BASE_US environment
 **     variable, or MACHINE_TIMER_BASE_US_DFL, as the base
 **     timer interval.
 **     (machine_spi_master_p): New function.
 **     (machine_spi_master_init): Implement on top of spi_init.
 **     (machine_spi_ready_p): New name for machine_spi_ready.
 **     Return early if spi_ready_p is set; set the variable to 1
 **     when returning true.  Send spi_buf_w only in slave mode
 **     (the variable is not used in master mode); call fds_write
 **     (with the spi_fds element the read came from) to do that
 **     (was: calling write directly.)
 **     (machine_spi_write): In master mode, set spi_ready_p to 0
 **     and call fds_write directly to broadcast the value given.
 **     (machine_uart_can_read_p): Ignore EINTR from read (was:
 **     closing the connection, as with the other error conditions.)
 **     (machine_uart_can_write_p): Call fds_accept before
 **     proceeding to check if any of the connections are ready
 **     for writing (was: after.)
 **     (port_setup): New unit-local function split off
 **     (machine_port_setup): this function.  Return early if the
 **     pointer is already initialized (was: file reopened, leaking fds.)
 **     (machine_port_setup): Also use PROT_READ for the write port,
 **     as we are using read-modify-write in machine_port_write anyway.
 **     (machine_port_setup): Use machine_port_map_t.
 **     (machine_port_read): Likewise.
 **     (machine_port_write): Likewise.
 **     (machine_port_read): Report tick.  Report state and masks as
 **     long long 16-digit hexadecimal values (were: long 8-digit),
 **     as that is the current width of the machine_port_map_t type.
 **     (machine_port_write): Likewise.
 **     (machine_port_write): Report but do not actually use port_w_mask
 **     (was: only the mask given); note that on AVR, writing 1 to a
 **     read-only port pin enables the internal pull-up resistor there.
 **
 ** 0.? 2020-03-14 10:15:12Z
 **     (sfn.bAiIg5Qewhua0_FToqidTK8htPDCSSBM5Isc3uLLkJw.c)
 */

/*** Code: */
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stddef.h>                     /* for ptrdiff_t */
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>                   /* for mmap */
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>                   /* for setitimer */
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>

#include <machine.h>

#define SIZE(a)  (sizeof (a) / sizeof (*(a)))

/** NB: this roughly corresponds to an AVR MCU running at 1048576 Hz */
#ifndef MACHINE_TIMER_BASE_US_DFL
#define MACHINE_TIMER_BASE_US_DFL   (244LL)
#endif

/** NB: set in machine_init */
static char *endpoint_prefix = 0;

static int
endpoint_filename (char *s, size_t z, const char *desc)
{
  assert (endpoint_prefix != 0);
  /* . */
  return
    snprintf (s, z, "%s%s", endpoint_prefix, desc);
}

static char *
endpoint_filename_alloc (const char *desc)
{
  assert (endpoint_prefix != 0);
  char *s;
  int r
    = asprintf (&s, "%s%s", endpoint_prefix, desc);
  /* . */
  return (r < 0 ? 0 : s);
}

static void
ensure_size (int fd, size_t size)
{
  off_t o = lseek (fd, 0, SEEK_END);
  if (o >= size) {
    /* . */
    return;
  }

  int r
    = ftruncate (fd, size);
  assert (r >= 0);

  /* . */
}

static int
new_socket (const char *desc)
{
  int fd
    = socket (AF_UNIX, SOCK_STREAM, 0);
  assert (fd >= 0);

  struct sockaddr_un addr;
  const size_t z = sizeof (addr);
  memset (&addr, 0, z);

  addr.sun_family = AF_UNIX;
  endpoint_filename (addr.sun_path, -1 + sizeof (addr.sun_path), desc);
  fprintf (stderr, "*** New socket (%s): %s\n", desc, addr.sun_path);

  int bl
    = bind (fd, (struct sockaddr *)&addr, z);
  assert (bl == 0);

  int rl
    = listen (fd, 3);
  assert (rl == 0);

  /** NB: ensure that SIGPIPE is ignored */
  static struct sigaction sa = {
    .sa_handler = SIG_IGN,
    .sa_flags = 0
  };
  sigemptyset (&(sa.sa_mask));
  sigaction (SIGPIPE, &sa, 0);

  /* . */
  return fd;
}

static int
fds_set (const int fds[], size_t size, fd_set *rp, fd_set *wp)
{
  int i, max;

  for (i = 0, max = -1; i < size; ++i) {
    if (fds[i] < 0) {
      continue;
    }

    if (wp != 0) {
      FD_SET (fds[i], wp);
    }
    if (rp != 0) {
      FD_SET (fds[i], rp);
    }
    if (fds[i] > max) {
      max = fds[i];
    }
  }

  /* . */
  return max;
}

static int
fds_accept (int server, int fds[], size_t size, int flags)
{
  int new
    = accept4 (server, 0, 0, flags);

  int i;
  for (i = 0; i < size && fds[i] >= 0; ++i)
    ;
  if (i >= size) {
    fputs ("Warning: closing connection: fds vector is full\n",
           stderr);
    close (new);
    /* . */
    return -1;
  }

  /* . */
  return fds[i] = new;
}

static void
fds_setup_x (const char *desc, int fds[], size_t size)
{
  assert (size > 0);
  if (fds[0] >= 0) {
    /** Already initialized, do nothing */
    /* . */
    return;
  }

  int fd
     = new_socket (desc);
  assert (fd >= 0);

  fds[0]
    = fd;
  int i;
  for (i = 1; i < size; ++i) {
    fds[i]
      = -1;
  }
  /* . */
}

static void
fds_write (int fds[], size_t size, char v)
{
  int i;
  for (i = 0; i < size; ++i) {
    if (fds[i] < 0) {
      continue;
    }

    ssize_t wr
      = write (fds[i], &v, 1);
    assert (wr <= 1);
    /** NB: silently ignore any write errors other than EPIPE */
    if (wr < 0 && errno == EPIPE) {
      close (fds[i]);
      fds[i] = -1;
    }
  }
  /* . */
}

void
machine_init (void)
{
  /** FIXME: we can optionally take argc, argv to use instead of getenv_r */

  if (0 != endpoint_prefix) {
    free  (endpoint_prefix);
  }
  /** Curiously enough, getenv_r is BSD-specific */
  char *ev;
  if (0 != (ev = getenv ("MACHINE_E_PREFIX"))) {
    endpoint_prefix
      = strdup (ev);
    assert (0 != endpoint_prefix);
  } else {
    ev = getenv ("TMPDIR") ? : "/tmp";
    int r
      = asprintf (&endpoint_prefix, "%s/%d.", ev, (int)getpid ());
    assert (r >= 0);
  }

  /* . */
}

static void
check_event (int or_wait_p, int nfds, fd_set *rp, fd_set *wp)
{
  struct timeval tv0 = {
    .tv_sec = 0, .tv_usec = 0
  };

  int r
    = select (nfds, rp, wp, 0, or_wait_p ? 0 : &tv0);
  /** NB: can return due to itimer */
  assert (r >= 0 || errno == EINTR);
  /* . */
}

static void
check_event_fds  (int or_wait_p, const int fds[], size_t size,
                  int *maxp, fd_set *rp, fd_set *wp)
{
  int fd_max = fds_set (fds, size, rp, wp);
  check_event (or_wait_p, 1 + fd_max, rp, wp);
  if (maxp != 0) {
    *maxp = fd_max;
  }
  /* . */
}

void
machine_wait (void)
{
  /** FIXME: rather, we should check the fds, and raise flags as needed */
  check_event (1, 0, 0, 0);
}


static int ticked_p = 0;
static unsigned long tick = 0;

static void
machine_timer_handler ()
{
  ticked_p = 1;
  ++tick;
}

void
machine_timer_setup (uint_least16_t divisor)
{
  unsigned long long in;
  char *ev, *endp;
  if (0 == (ev = getenv ("MACHINE_TIMER_BASE_US"))
      || (in = strtoull (ev, &endp, 0),
          endp <= ev || *endp != 0)) {
    in = MACHINE_TIMER_BASE_US_DFL;
  }

  /** NB: frequency divisor is interval multiplier */
  in *= divisor;
  struct itimerval it = {
    .it_interval  = {
      .tv_sec   = in / 1000000ULL,
      .tv_usec  = in % 1000000ULL
    },
    .it_value     = { .tv_sec = 0, .tv_usec = 1 }
  };
  /** FIXME:  we cannot use SA_RESTART as we need it to interrupt select
   **
   ** The problem is that the signal can thus interrupt other calls,
   ** which we do not consistently handle now.
   */
  struct sigaction sa = {
    .sa_handler = &machine_timer_handler,
    .sa_flags = 0,
  };
  sigemptyset (&(sa.sa_mask));
  sigaction (SIGALRM, &sa, 0);

  setitimer (ITIMER_REAL, &it, 0);
  /* . */
}

int_least8_t
machine_timer_ticked_p (void)
{
  int r = ticked_p;
  ticked_p = 0;
  /* . */
  return r;
}


#if 0
/** FIXME: not implemented yet */
uint_least8_t machine_eeprom_read (ptrdiff_t);
void machine_eeprom_write (ptrdiff_t, uint_least8_t);
#endif


/** FIXME: support arbitrary number of SPI channels? */
static int spi_fds[64] = { -1 };
int_least8_t spi_master_p = 0, spi_ready_p = 0;

static void
spi_init (void)
{
  fds_setup_x ("spi", spi_fds, SIZE (spi_fds));
  /* . */
}

int_least8_t
machine_spi_master_p (uint_least8_t cc)
{
  assert (0 == cc);
  /* . */
  return spi_master_p;
}

void
machine_spi_master_init (uint_least8_t cc)
{
  assert (0 == cc);
  spi_init ();
  spi_master_p = spi_ready_p = 1;
}

void
machine_spi_slave_init (uint_least8_t cc)
{
  assert (0 == cc);
  spi_init ();
  spi_master_p = spi_ready_p = 0;
}

static char spi_buf_w = 0, spi_buf_r = 0;

/** The code below implements a kind of multiplexing
 ** on the SPI sockets as follows.
 **
 ** When in master mode, the octet sent is broadcast to all
 ** the clients (if any.)  The first response we see, we store
 ** in spi_buf_r; spi_buf_w is not used.
 **
 ** Otherwise, we wait until we can receive an octet.  We then
 ** store it in spi_buf_r, and immediately respond with the
 ** one stored in spi_buf_w.
 */

int_least8_t
machine_spi_ready_p (uint_least8_t cc)
{
  assert (0 == cc);
  if (0 != spi_ready_p) {
    /* . */
    return spi_ready_p;
  }
  assert (spi_fds[0] >= 0);

  fd_set rs;
  check_event_fds (0, spi_fds, SIZE (spi_fds), 0, &rs, 0);

  if (FD_ISSET (spi_fds[0], &rs)) {
    fds_accept (spi_fds[0], 1 + spi_fds, -1 + SIZE (spi_fds),
                SOCK_NONBLOCK);
  }

  int i;
  for (i = 1; i < SIZE (spi_fds); ++i) {
    int fd
      = spi_fds[i];
    if (fd < 0 || ! FD_ISSET (fd, &rs)) {
      continue;
    }

    ssize_t rd
      = read  (fd, &spi_buf_r, 1);
    assert (rd <= 1);
    if (rd < 0) {
      close (fd);
      spi_fds[i] = -1;
      continue;
    }

    if (! spi_master_p) {
      fds_write (i + spi_fds, 1, spi_buf_w);
    }

    /* . */
    return spi_ready_p = 1;
  }

  /* . */
  return 0;
}

uint_least8_t
machine_spi_read (uint_least8_t cc)
{
  /* . */
  return spi_buf_r;
}

void
machine_spi_write (uint_least8_t cc, uint_least8_t v)
{
  if (machine_spi_master_p (cc)) {
    spi_ready_p = 0;
    fds_write (1 + spi_fds, -1 + SIZE (spi_fds), v);
    /* . */
    return;
  }
  spi_buf_w = v;
  /* . */
}


/** FIXME: support arbitrary number of UART channels? */
static int uart_fds[64] = { -1 };

void
machine_uart_setup (uint_least8_t cc)
{
  assert (0 == cc);
  fds_setup_x ("uart", uart_fds, SIZE (uart_fds));
}

static int_least16_t uart_buf_r = -1;

int_least8_t
machine_uart_can_read_p (uint_least8_t cc)
{
  assert (0 == cc);
  assert (uart_fds[0] >= 0);

  if (uart_buf_r >= 0) {
    /* . */
    return 1;
  }

  fd_set rs;
  check_event_fds (0, uart_fds, SIZE (uart_fds), 0, &rs, 0);

  if (FD_ISSET (uart_fds[0], &rs)) {
    fds_accept (uart_fds[0], 1 + uart_fds, -1 + SIZE (uart_fds),
                SOCK_NONBLOCK);
  }

  int i;
  for (i = 1; i < SIZE (uart_fds); ++i) {
    int fd
      = uart_fds[i];
    if (fd < 0 || ! FD_ISSET (fd, &rs)) {
      continue;
    }

    char rc;
    ssize_t rd
      = read (fd, &rc, 1);
    assert (rd <= 1);

    if (rd > 0) {
      uart_buf_r = rc;
      /* . */
      return 1;
    } else if (errno == EINTR) {
      continue;
    }

    close (fd);
    uart_fds[i] = -1;
  }

  /* . */
  return 0;
}

int_least8_t
machine_uart_can_write_p (uint_least8_t cc)
{
  assert (0 == cc);
  assert (uart_fds[0] >= 0);

  fd_set rs;
  check_event_fds (0, uart_fds, 1, 0, &rs, 0);

  if (FD_ISSET (uart_fds[0], &rs)) {
    fds_accept (uart_fds[0], 1 + uart_fds, -1 + SIZE (uart_fds),
                SOCK_NONBLOCK);
  }

  fd_set ws;
  check_event_fds (0, 1 + uart_fds, -1 + SIZE (uart_fds), 0, 0, &ws);

  int i;
  for (i = 1; i < SIZE (uart_fds); ++i) {
    int fd
      = uart_fds[i];
    if (fd >= 0 && FD_ISSET (fd, &ws)) {
      /* . */
      return 1;
    }
  }

  /* . */
  return 0;
}

int_least16_t
machine_uart_read (uint_least8_t cc)
{
  int_least16_t r = uart_buf_r;
  uart_buf_r = -1;
  /* . */
  return r;
}

void
machine_uart_write (uint_least8_t cc, uint_least8_t v)
{
  assert (0 == cc);
  fds_write (1 + uart_fds, -1 + SIZE (uart_fds), v);
  /* . */
}


/** NB: port_rp, port_wp are mmaped; write to file to modify *port_rp */
static machine_port_map_t *port_rp = 0, *port_wp = 0;
static machine_port_map_t port_r_mask = 0, port_w_mask = 0;
static machine_port_map_t port_r_save = (~ 0);

static void
port_setup (const char *desc,
            machine_port_map_t **portp, size_t size, int protection)
{
  if (0 != *portp) {
    /* . */
    return;
  }

  char *fn
    = endpoint_filename_alloc (desc);
  assert (fn != 0);
  /** NB: we need read access as mmap can EACCESS for PROT_WRITE otherwise */
  /** NB: we need write access too, to be able to ensure needed size */
  int fd = open (fn, O_RDWR | O_CREAT, 0666);
  assert (fd >= 0);
  fprintf (stderr, "*** New %s: %s\n", desc, fn);
  free (fn);
  ensure_size (fd, size);
  void *mm
    = mmap (0, size, protection, MAP_SHARED, fd, 0);
  assert (mm != MAP_FAILED);
  *portp = mm;
}

void
machine_port_setup (machine_port_map_t read_mask,
                    machine_port_map_t write_mask)
{
  if (0 != (port_r_mask = read_mask)) {
    port_setup ("port-r", &port_rp, sizeof (*port_rp), PROT_READ);
  }
  if (0 != (port_w_mask = write_mask)) {
    port_setup ("port-w", &port_wp, sizeof (*port_wp),
                  (PROT_WRITE | PROT_READ));
  }
  /* . */
}

machine_port_map_t
machine_port_read (uint_least8_t shift, machine_port_map_t mask)
{
  assert (0 != port_rp);
  machine_port_map_t new = *port_rp;
  if (((new ^ port_r_save) & port_r_mask)) {
    port_r_save = new;
    fprintf (stderr,
             ("*** @%ld PORT[r]: state = %016llx,"
              " shift = %d, mask = %016llx & %016llx\n"),
             tick, (long long)new, (int)shift,
             (long long)mask, (long long)port_r_mask);
  }
  /* . */
  return ((new >> shift) & mask);
}

void
machine_port_write (machine_port_map_t data, machine_port_map_t mask)
{
  /** NB: we are doing read-modify-write here, hence PROT_READ above */
  machine_port_map_t old = *port_wp;
  machine_port_map_t new = ((old & (~ mask)) | (data & mask));
  if (new == old) {
    /* . */
    return;
  }
  *port_wp = new;
  fprintf (stderr,
           ("*** @%ld PORT[w]: state = %016llx,"
            " data = %016llx, mask = %016llx & %016llx\n"),
           tick, (long long)new, (long long)data,
           (long long)mask, (long long)port_w_mask);
}

/*** machine.c ends here */
